"""Module containing custom exceptions."""

from typing import Any


class BIDSQueryError(Exception):
    """
    BIDS File Querying Exception.

    Raised in ``TimeseriesExtractor.get_bold()`` when a pybids ``BIDSLayout`` returns no subject IDs when querying the
    BIDS directory.

    This error may occur due to:
      - Incorrect template space (e.g. using the default "MNI152NLin2009cAsym" when a different space is used).
      - File naming issues where required entities (e.g. "sub-", "space-", "task-", "desc-") are missing.
      - An incorrect task name specified in the `task` parameter.
      - The directory being changed during the current Python session (e.g. new files added, file names changed, etc),\
      resulting in the cache needing to be cleared using ``TimeseriesExtractor._call_layout.cache_clear()``.

    Refer to `NeuroCAPs' BIDS Structure and Entities <https://neurocaps.readthedocs.io/en/stable/bids.html>`_ for
    additional information on the expected directory structure and entities needed for querying.
    """

    def __init__(self, message: Any) -> None:
        super().__init__(message)


class NoElbowDetectedError(Exception):
    """
    Elbow Method Failure Exception.

    This exception occurs in ``CAP.get_caps()`` when ``cluster_selection_method`` is set to "elbow" but kneed's
    ``KneeLocator`` fails to detect a point of maximum curvature in the elbow curve.

    This error may occur due to:
      - The range of tested cluster sizes (k) is too restrictive and should be expanded, as the elbow likely occurs at\
      a larger cluster size.
      - Multiple elbows existing; thus, the value of the sensitivity (``S``) parameter of ``KneeLocator`` should be\
      increased to be more conservative.
      - The data lacking a natural clustering structure (exceptionally rare for fMRI data).

    .. versionadded:: 0.24.0
    """

    def __init__(self, message: Any) -> None:
        super().__init__(message)



from .capgetter import _CAPGetter
from .cap2statmap import _build_tree, _cap2statmap, _get_target_indices
from .run_kmeans import _run_kmeans
from .heatmap_utils import _create_display, _save_contents



"""Function for averaging subject-level transition probabilities and producing visualizations"""

from typing import Optional, Union

import matplotlib.pyplot as plt, pandas as pd

from .._utils import _PlotDefaults, _check_kwargs, _create_display, _logger, _save_contents

LG = _logger(__name__)


def transition_matrix(
    trans_dict: dict[str, pd.DataFrame],
    output_dir: Optional[str] = None,
    suffix_title: Optional[str] = None,
    suffix_filename: Optional[str] = None,
    show_figs: bool = True,
    save_plots: bool = True,
    return_df: bool = True,
    save_df: bool = True,
    **kwargs,
) -> Union[pd.DataFrame, None]:
    """
    Generate and Visualize the Averaged Transition Probabilities.

    Averages subject-level transition probabilities to produce a transition probability matrix. One matrix is
    generated per group.

    Parameters
    ----------
    trans_dict: :obj: `dict[str, pd.DataFrame]`
        A dictionary mapping groups to pandas DataFrame containing the transition probabilities for each subject.
        This assumes the output from ``CAP.calculate_metrics`` is being used, specifically
        ``metrics_output["transition_probability"]``.

    output_dir: :obj:`str` or :obj:`None`, default=None
        Directory to save plots (if ``save_plots`` is True) and transition probability matrices DataFrames (if
        ``save_df`` is True) to. The directory will be created if it does not exist. Plots and dataframes will not
        be saved if None.

    suffix_title: :obj:`str` or :obj:`None`, default=None
        Appended to the title of each plot.

    suffix_filename: :obj:`str` or :obj:`None`, default=None
        Appended to the filename of each saved plot if ``output_dir`` is provided.

    show_figs: :obj:`bool`, default=True
        Display figures.

    save_plots: :obj:`bool`, default=True
        If True, plots are saves as png images. For this to be used, ``output_dir`` must be specified.

    return_df: :obj:`bool`, default=False
        If True, returns a dictionary with a transition probability matrix for each group.

    save_df: :obj:`bool`, default=False,
        If True, saves the transition probability matrix contained in the DataFrames as csv files. For this to be used,
        ``output_dir`` must be specified.

    **kwargs
        Keyword arguments used when modifying figures. Valid keywords include:

        - dpi: :obj:`int`, default=300 -- Dots per inch for the figure.
        - figsize: :obj:`tuple`, default=(8, 6) -- Size of the figure in inches.
        - fontsize: :obj:`int`, default=14 -- Font size for the plot title, x-axis title, and y-axis title of each plot.
        - xticklabels_size: :obj:`int`, default=8 -- Font size for x-axis tick labels.
        - yticklabels_size: :obj:`int`, default=8 -- Font size for y-axis tick labels.
        - shrink: :obj:`float`, default=0.8 -- Fraction by which to shrink the colorbar.
        - cbarlabels_size: :obj:`int`, default=8 -- Font size for the colorbar labels.
        - xlabel_rotation: :obj:`int`, default=0 -- Rotation angle for x-axis labels.
        - ylabel_rotation: :obj:`int`, default=0 -- Rotation angle for y-axis labels.
        - annot: :obj:`bool`, default=False -- Add values to each cell.
        - annot_kws: :obj:`dict`, default=None, -- Customize the annotations.
        - fmt: :obj:`str`, default=".2g" -- Modify how the annotated vales are presented.
        - linewidths: :obj:`float`, default=0 -- Padding between each cell in the plot.
        - borderwidths: :obj:`float`, default=0 -- Width of the border around the plot.
        - linecolor: :obj:`str`, default="black" -- Color of the line that separates each cell.
        - edgecolors: :obj:`str` or :obj:`None`, default=None -- Color of the edges.
        - alpha: :obj:`float` or :obj:`None`, default=None -- Controls transparency and ranges from 0 (transparent) to 1 (opaque).
        - bbox_inches: :obj:`str` or :obj:`None`, default="tight" -- Alters size of the whitespace in the saved image.
        - cmap: :obj:`str`, :obj:`callable` default="coolwarm" -- Color map for the plot cells. Options include\
            strings to call seaborn's pre-made palettes, ``seaborn.diverging_palette`` function to generate custom\
            palettes, and ``matplotlib.color.LinearSegmentedColormap`` to generate custom palettes.
        - vmin: :obj:`float` or :obj:`None`, default=None -- The minimum value to display in colormap.
        - vmax: :obj:`float` or :obj:`None`, default=None -- The maximum value to display in colormap.

    Returns
    -------
    dict[str, pd.DataFrame]
        An instance of a pandas DataFrame for each group if ``return_df`` is True.

    Note
    ----
    **Dataframe Representation**: Rows represent "from" and columns represent "to". For instance,
    the probability at ``df.loc["CAP-1", "CAP-2"]`` represents the averaged probability from transitioning from
    CAP-1 to CAP-2.

    +------------+---------+-------+-------+
    | From/To    |  CAP-1  | CAP-2 | CAP-3 |
    +============+=========+=======+=======+
    | CAP-1      |  0.40   | 0.35  | 0.25  |
    +------------+---------+-------+-------+
    | CAP-2      |  0.20   | 0.45  | 0.35  |
    +------------+---------+-------+-------+
    | CAP-3      |  0.35   | 0.18  |  0.47 |
    +------------+---------+-------+-------+
    """
    assert isinstance(trans_dict, dict), "transition_dict must be in the form dict[str, pd.DataFrame]."

    if suffix_filename is not None and output_dir is None:
        LG.warning("`suffix_filename` supplied but no `output_dir` specified. Files will not be saved.")

    # Create plot dictionary
    plot_dict = _check_kwargs(_PlotDefaults.transition_matrix(), **kwargs)

    trans_mat_dict = {}

    for group in trans_dict:
        df = trans_dict[group]
        # Get indices and averaged probabilities
        indices, averaged_probabilities = df.iloc[:, 3:].mean().index, df.iloc[:, 3:].mean().values
        # Get the maximum CAP
        max_cap = str(max([float(i) for i in indices])).split(".")[0]
        cap_names = [f"CAP-{num}" for num in range(1, int(max_cap) + 1)]
        trans_mat = pd.DataFrame(index=cap_names, columns=cap_names, dtype="float64")
        # Add name to index
        trans_mat.index.name = "From/To"

        # Create matrix
        for location, name in enumerate(indices):
            trans_mat.loc[f"CAP-{name.split('.')[0]}", f"CAP-{name.split('.')[1]}"] = averaged_probabilities[location]

        display = _create_display(trans_mat, plot_dict, suffix_title, group, "trans")

        # Store df in dict
        trans_mat_dict[group] = trans_mat
        # Save figure & dataframe
        if output_dir:
            _save_contents(
                output_dir,
                suffix_filename,
                group,
                trans_mat_dict,
                plot_dict,
                save_plots,
                save_df,
                display,
                call="trans",
            )

        # Display figures
        plt.show() if show_figs else plt.close()

    if return_df:
        return trans_mat_dict



"""Function for changing the dtype of timeseries data."""

import copy
from typing import Union, Optional

import numpy as np

from .._utils import _convert_pickle_to_dict, _dicts_to_pickles, _logger
from ..typing import SubjectTimeseries

LG = _logger(__name__)


def change_dtype(
    subject_timeseries_list: Union[list[SubjectTimeseries], list[str]],
    dtype: Union[str, np.floating],
    return_dicts: bool = True,
    output_dir: Optional[str] = None,
    filenames: Optional[list[str]] = None,
) -> Union[dict[str, SubjectTimeseries], None]:
    """
    Perform Participant-wise Dtype Conversion.

    Changes the dtypes of each participants NumPy array.

    Parameters
    ----------
    subject_timeseries_list: :obj:`list[dict[str, dict[str, np.ndarray]]]` or :obj:`list[str]`
        A list where each element consist of a dictionary mapping subject IDs to their run IDs and associated
        timeseries (TRs x ROIs) as a NumPy array. Can also be a list consisting of paths to pickle files
        containing this same structure. Refer to documentation for ``SubjectTimeseries`` in the "See Also" section for
        an example structure.

    dtype: :obj:`bool` or :obj:`np.floating`
        Target data type (e.g "float32" or ``np.float32``) to convert each participant's NumPy arrays into.

    return_dicts: :obj:`bool`, default=True
        If True, returns a single dictionary containing the converted input dictionaries. Keys are named "dict_{0}"
        where {0} corresponds to the dictionary's position in the input list.

    output_dir: :obj:`str` or :obj:`None`, default=None
        Directory to save the converted ``subject_timeseries`` as pickle files. The directory will be created if it
        does not exist. Dictionaries will not be saved if None.

    filenames: :obj:`list[str]` or :obj:`None`, default=None
        A list of names to save the dictionaries with changed dtypes as. Names are matched to dictionaries by position
        (e.g., a file name in the 0th position will be the file name for the dictionary in the 0th position of
        ``subject_timeseries_list``). If None and ``output_dir`` is specified, uses default file names -
        "subject_timeseries_{0}_float{1}.pkl" (where {0} indicates the original input order and {1} is the dtype.

    Returns
    -------
    dict[str, SubjectTimeseries]
        A nested dictionary containing the converted subject timeseries if ``return_dicts`` is True.

    See Also
    --------
    :data:`neurocaps.typing.SubjectTimeseries`
        Type definition for the subject timeseries dictionary structure.

    Warning
    -------
    **Floating Point Precision**: The minimum recommended floating-point dtype is *float32*, as lower precision may
    introduce rounding errors that affect calculations.
    """
    assert isinstance(subject_timeseries_list, list), "`subject_timeseries_list` must be a list."

    if filenames is not None and output_dir is None:
        LG.warning("`filenames` supplied but no `output_dir` specified. Files will not be saved.")

    changed_dtype_dicts = {}

    for indx, curr_dict in enumerate(subject_timeseries_list):
        if isinstance(curr_dict, str) and curr_dict.endswith(".pkl"):
            curr_dict = _convert_pickle_to_dict(curr_dict)
        else:
            curr_dict = copy.deepcopy(curr_dict)

        for subj_id in curr_dict:
            for run in curr_dict[subj_id]:
                curr_dict[subj_id][run] = curr_dict[subj_id][run].astype(dtype)

        changed_dtype_dicts[f"dict_{indx}"] = curr_dict

    if output_dir:
        _dicts_to_pickles(
            output_dir=output_dir, dict_list=changed_dtype_dicts, filenames=filenames, call="change_dtype"
        )

    if return_dicts:
        return changed_dtype_dicts
